# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/07_training.metrics.ipynb (unless otherwise specified).

__all__ = ['rpd', 'rpiq', 'stb', 'mape', 'lccc', 'eval_reg']

# Cell
#nbdev_comment from __future__ import annotations
import numpy as np
from sklearn.metrics import (r2_score, mean_squared_error, mean_absolute_error)
from fastcore.test import *

# Cell
def rpd(y:np.ndarray, # Target true value
        y_hat:np.ndarray, # Target predicted value
       ):
    """Ratio of Performance to Deviation"""
    sd = np.std(y, ddof=1)
    sep = np.sqrt(mean_squared_error(y, y_hat))
    return sd / sep

# Cell
def rpiq(y:np.ndarray, # Target true value
         y_hat:np.ndarray, # Target predicted value
        ):
    """Ratio of Performance to Inter-Quartile"""
    q1, q3 = np.percentile(y, [25, 75])
    sep = np.sqrt(mean_squared_error(y, y_hat))
    return (q3 - q1) / sep

# Cell
def stb(y:np.ndarray, # Target true value
        y_hat:np.ndarray, # Target predicted value
       ):
    """Standardized Bias"""
    q1, q3 = np.percentile(y, [25, 75])
    _bias = np.mean(y - y_hat)
    return _bias / (q3 - q1)

# Cell
def mape(y:np.ndarray, # Target true value
         y_hat:np.ndarray, # Target predicted value
        ):
    """Mean Absolute Percentage Error"""
    return 100 * np.mean(np.abs(y - y_hat) / np.abs(y))

# Cell
def lccc(y:np.ndarray, # Target true value
         y_hat:np.ndarray, # Target predicted value
        ):
    """Lin’s concordance correlation coefficient"""
    y = y.flatten()
    y_hat = y_hat.flatten()
    numerator = 2 * np.corrcoef(y, y_hat)[0, 1] * np.std(y) * np.std(y_hat)
    denominator = np.var(y) + np.var(y_hat) + (np.mean(y) - np.mean(y_hat))**2
    return numerator / denominator

# Cell
def eval_reg(y:np.ndarray, # Target true value
             y_hat:np.ndarray, # Target predicted value
             is_log:bool=True, # True if evaluated values are log-10 transformed
            ):
    """Return metrics bundle (rpd, rpiq, r2, lccc, rmse, mse, mae, mape, bias, stb)"""
    _rpd = rpd(y, y_hat)
    _rpiq = rpiq(y, y_hat)
    _r2_score = r2_score(y, y_hat)
    _bias = np.mean(y - y_hat)
    _stb = stb(y, y_hat)
    _lccc = lccc(y, y_hat)
    if is_log:
        y = np.power(10, y)
        y_hat = np.power(10, y_hat)
    _rmse = np.sqrt(mean_squared_error(y, y_hat))
    _mae = mean_absolute_error(y, y_hat)
    _mape = mape(y, y_hat)
    return {
        'rpd': _rpd,
        'rpiq': _rpiq,
        'r2': _r2_score,
        'lccc': _lccc,
        'rmse': _rmse,
        'mse': mean_squared_error(y, y_hat),
        'mae': _mae,
        'mape': _mape,
        'bias': _bias,
        'stb': _stb
    }


# AUTOGENERATED BY NBDEV! DO NOT EDIT!

__all__ = ["index", "modules", "custom_doc_links", "git_url"]

index = {"say_hello": "00_core.ipynb",
         "load_kssl": "01_data.loading.ipynb",
         "PRIMARY_COLOR": "03_vis.core.ipynb",
         "ACCENT_COLOR": "03_vis.core.ipynb",
         "DEFAULT_STYLE": "03_vis.core.ipynb",
         "centimeter": "03_vis.core.ipynb",
         "set_style": "03_vis.core.ipynb",
         "plot_spectra": "03_vis.core.ipynb",
         "summary_plot": "03_vis.core.ipynb",
         "plot_validation_curve": "03_vis.core.ipynb",
         "plot_learning_curve": "03_vis.core.ipynb",
         "plot_capacity": "03_vis.core.ipynb",
         "select_y": "04_data.selection.ipynb",
         "select_tax_order": "04_data.selection.ipynb",
         "select_X": "04_data.selection.ipynb",
         "get_y_by_order": "04_data.selection.ipynb",
         "log_transform_y": "05_data.transform.ipynb",
         "CO2_REGION": "05_data.transform.ipynb",
         "TakeDerivative": "05_data.transform.ipynb",
         "SNV": "05_data.transform.ipynb",
         "DropSpectralRegions": "05_data.transform.ipynb",
         "rpd": "07_training.metrics.ipynb",
         "rpiq": "07_training.metrics.ipynb",
         "stb": "07_training.metrics.ipynb",
         "mape": "07_training.metrics.ipynb",
         "lccc": "07_training.metrics.ipynb",
         "eval_reg": "07_training.metrics.ipynb",
         "compute_valid_curve": "10_training.plsr.ipynb",
         "PLS_model": "10_training.plsr.ipynb",
         "Evaluator": "12_training.cnn.ipynb",
         "Model": "12_training.cnn.ipynb",
         "weights_init": "12_training.cnn.ipynb",
         "Learner": "12_training.cnn.ipynb",
         "SpectralDataset": "13_data.torch.ipynb",
         "DataLoaders": "13_data.torch.ipynb",
         "SNV_transform": "13_data.torch.ipynb",
         "Noop": "13_data.torch.ipynb"}

modules = ["core.py",
           "data/loading.py",
           "vis/core.py",
           "data/selection.py",
           "data/transform.py",
           "training/metrics.py",
           "training/plsr.py",
           "training/cnn.py",
           "data/torch.py"]

doc_url = "https://franckalbinet.github.io/mirzai/"

git_url = "https://github.com/franckalbinet/mirzai/tree/master/"

def custom_doc_links(name): return None



# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/05_data.transform.ipynb (unless otherwise specified).

__all__ = ['log_transform_y', 'CO2_REGION', 'TakeDerivative', 'SNV', 'DropSpectralRegions']

# Cell
#nbdev_comment from __future__ import annotations
import numpy as np
from sklearn.base import BaseEstimator, TransformerMixin
from scipy.signal import savgol_filter
from fastcore.test import *

from .loading import load_kssl
from ..vis.core import plot_spectra

# Cell
def log_transform_y(data:tuple # (X, y, X_id, depth_order)
                   ):
    """
    Log-10 transform of the target value

    Takes and returns all (X, y, X_id, depth_order) tuple to be able to pipe the function
    """
    X, y, X_id, depth_order = data
    y = np.log10(y)
    return X, y, X_id, depth_order

# Cell
CO2_REGION = [2389,  2269]

# Cell
class TakeDerivative(BaseEstimator, TransformerMixin):
    """Creates scikit-learn derivation custom transformer

    Args:
        window_length: int, optional
            Specify savgol filter smoothing window length

        polyorder: int, optional
            Specify order of the polynom used to interpolate derived signal

        deriv: int, optional
            Specify derivation degree

    Returns:
        scikit-learn custom transformer
    """
    def __init__(self, window_length=11, polyorder=1, deriv=1):
        self.window_length = window_length
        self.polyorder = polyorder
        self.deriv = deriv

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        return savgol_filter(X, self.window_length, self.polyorder, self.deriv)

# Cell
class SNV(BaseEstimator, TransformerMixin):
    """Creates scikit-learn SNV custom transformer

    Args:
        None

    Returns:
        scikit-learn custom transformer
    """
    def __init__(self):
        pass

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        mean, std = np.mean(X, axis=1).reshape(-1, 1), np.std(X, axis=1).reshape(-1, 1)
        return (X - mean)/std

# Cell
class DropSpectralRegions(BaseEstimator, TransformerMixin):
    """Creates scikit-learn custom transformer dropping specific spectral region(s)

    Args:
        wavenumbers: list
            List of wavenumbers where absorbance measured

        regions: list
            List of region(s) to drop

    Returns:
        scikit-learn custom transformer
    """
    def __init__(self, wavenumbers, regions=[2389,  2269]):
        self.wavenumbers = wavenumbers
        self.regions = regions

    def _sanitize(self, regions):
        nb_regions = len(np.array(regions).shape)
        return np.array([regions]) if nb_regions == 1 else np.array(regions)

    def _exists(self, wavenumbers, regions):
        for wn in regions.flatten():
            assert wn in wavenumbers, 'Wavenumber "{}" does not exist'.format(wn)

    def fit(self, X, y=None):
        return self

    def transform(self, X, y=None):
        regions = self._sanitize(self.regions)
        X_transformed = np.copy(X)
        self._exists(self.wavenumbers, regions)
        for region in regions:
            high, low = region
            mask = (self.wavenumbers <= high) & (self.wavenumbers >= low)
            X_transformed[:, mask] = 0

        return X_transformed


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/01_data.core.ipynb (unless otherwise specified).

__all__ = ['load_file']

# Cell
import numpy as np
import os
from fastcore.test import *

# Cell
def load_file():
    return np.array([1, 2, 3])

# Cell
# def load_file(name, path='./data', loader=np.load):
#     """
#     Loads individual data file.

#     Args:
#         name: the name of the file
#         path: the path containing
#         data
#         loader: the function used to load. E.g. `np.load` or `pickle.load`

#     Returns:
#         data as numpy.ndarray or Dict (when pickle is used)
#     """

#     with open(os.path.join(path, name), 'rb') as f:
#         return loader(f)


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/10_training.plsr.ipynb (unless otherwise specified).

__all__ = ['compute_valid_curve', 'PLS_model', 'Evaluator']

# Cell
#nbdev_comment from __future__ import annotations

# Python utils
from collections import OrderedDict
from tqdm.auto import tqdm

# mirzai utils
from ..data.loading import load_kssl
from ..data.selection import (select_y, select_tax_order, select_X)
from ..data.transform import (log_transform_y, SNV, TakeDerivative,
                                   DropSpectralRegions, CO2_REGION)
from .metrics import eval_reg

# Data science stack
import numpy as np
from sklearn.model_selection import train_test_split
from sklearn.pipeline import Pipeline
from sklearn.cross_decomposition import PLSRegression
from sklearn.metrics import mean_squared_error

from fastcore.test import *
from fastcore.transform import compose

# Cell
def compute_valid_curve(X:np.ndarray, # Spectra with shape (n_samples, n_wavenumbers)
                        y:np.ndarray, # Target with shape (n_samples)
                        X_names:np.ndarray, # Wavenumbers name with shape (n_wavenumbers)
                        mask:np.ndarray=None, # Mask (e.g for specific Soil Taxonomy Orders) with shape (n_samples)
                        pls_components:list=range(5, 120, 5), # List of PLSR components to try
                        seeds:list=range(20), # List of random seeds to use for multiple train/test splits
                        test_size:float=0.1 # Train/test ratio
                       ):
    "Train the PLSR model on training & evaluate it on the valid set as # pls components increases"
    history = OrderedDict({'pls_components': pls_components,
                           'train_score': [],
                           'valid_score': []})
    for seed in tqdm(seeds):
        if mask is None:
            mask = np.full(len(X), True)
        X_train, X_valid, y_train, y_valid = train_test_split(X[mask, : ], y[mask],
                                                              test_size=test_size,
                                                              random_state=seed)

        train_score = []
        valid_score = []
        for cpts in tqdm(pls_components):
            pipe = Pipeline([('snv', SNV()),
                            ('derivative', TakeDerivative(window_length=11, polyorder=1)),
                            ('dropper', DropSpectralRegions(X_names, regions=CO2_REGION)),
                            ('model', PLSRegression(n_components=cpts))])

            pipe.fit(X_train, y_train)

            train_score.append(mean_squared_error(pipe.predict(X_train), y_train))
            valid_score.append(mean_squared_error(pipe.predict(X_valid), y_valid))

        history['train_score'].append(train_score)
        history['valid_score'].append(valid_score)

    return history

# Cell
class PLS_model():
    "Partial Least Squares model runner"
    def __init__(self, X_names, pipeline_kwargs={}):
        self.X_names = X_names
        self.pipeline_kwargs = pipeline_kwargs
        self.model = None

    def fit(self, data):
        X, y = data
        self.model = Pipeline([
            ('snv', SNV()),
            ('derivative', TakeDerivative(**self.pipeline_kwargs['derivative'])),
            ('dropper', DropSpectralRegions(self.X_names, **self.pipeline_kwargs['dropper'])),
            ('model', PLSRegression(**self.pipeline_kwargs['model']))])
        self.model.fit(X, y)
        return self

    def predict(self, data):
        X, y = data
        return (self.model.predict(X), y)

    def eval(self, data, is_log=True):
        X, y = data
        return eval_reg(y, self.model.predict(X))

# Cell
class Evaluator():
    def __init__(self, data, depth_order, X_names,
                 seeds=range(20), pipeline_kwargs={}, split_ratio=0.1):
        self.seeds = seeds
        self.X, self.y = data
        self.X_names = X_names
        self.depth_order = depth_order
        self.split_ratio = split_ratio
        self.pipeline_kwargs = pipeline_kwargs
        self.models = []
        self.perfs = OrderedDict({'train': [], 'test': []})

    def train_multiple(self):
        for seed in tqdm(self.seeds):
            X_train, X_test, y_train, y_test, depth_order_train, depth_order_test = self._splitter(seed)
            model = PLS_model(self.X_names, self.pipeline_kwargs)
            model.fit((X_train, y_train))
            self.models.append(model)

    def eval_on_train(self, reducer):
        perfs = []
        for i, seed in enumerate(self.seeds):
            X_train, X_test, y_train, y_test, _, _ = self._splitter(seed)
            perf = self.models[i].eval((X_train, y_train))
            perf['n'] = len(X_train)
            perfs.append(perf)
        if reducer:
            perfs = self.reduce(perfs, reducer)
        return perfs

    def eval_on_test(self, order=-1, reducer=None):
        perfs = []
        for i, seed in tqdm(enumerate(self.seeds)):
            X_train, X_test, y_train, y_test, depth_order_train, depth_order_test = self._splitter(seed)
            if order != - 1:
                mask = depth_order_test[:, 1] == order
                X_test, y_test = X_test[mask, :], y_test[mask]
            perf = self.models[i].eval((X_test, y_test))
            perf['n'] = len(X_test)
            perfs.append(perf)
        if reducer:
            perfs = self.reduce(perfs, reducer)
        return perfs

    def _splitter(self, seed):
        return train_test_split(self.X, self.y, self.depth_order,
                                test_size=self.split_ratio,
                                random_state=seed)

    def reduce(self, perfs, fn=np.mean):
        results = {}
        for metric in perfs[0].keys():
            result = fn(np.array([perf[metric] for perf in perfs]))
            results[metric] = result
        return results


# AUTOGENERATED! DO NOT EDIT! File to edit: nbs/12_training.cnn.ipynb (unless otherwise specified).

__all__ = ['Model', 'weights_init', 'Learner', 'Evaluator']

# Cell
#nbdev_comment from __future__ import annotations

# Python utils
from collections import OrderedDict
from tqdm.auto import tqdm
import functools
import operator
from pathlib import Path

import numpy as np

from sklearn.model_selection import train_test_split

from ..data.loading import load_kssl
from ..data.selection import (select_y, select_tax_order, select_X)
from ..data.transform import log_transform_y
from ..data.torch import DataLoaders, SNV_transform
from .metrics import eval_reg

# Deep Learning stack
import torch
from torch import nn
from torchvision import transforms
from torch.utils.data import DataLoader, Dataset
from torch.optim import Adam
from torch.optim.lr_scheduler import CyclicLR
from torch.nn import MSELoss

from fastcore.test import *
from fastcore.basics import store_attr
from fastcore.transform import compose

# Cell
# https://datascience.stackexchange.com/questions/40906/determining-size-of-fc-layer-after-conv-layer-in-pytorch
class Model(nn.Module):
    def __init__(self, input_dim, in_channel=1, out_channel=16, is_classifier=False,
                 dropout=0.4):
        super(Model, self).__init__()
        # Build the neural network
        self.feature_extractor = nn.Sequential(
            self.make_convpool_block(in_channel, out_channel),
            self.make_convpool_block(out_channel, out_channel*2),
            self.make_convpool_block(out_channel*2, out_channel*4),
            self.make_convpool_block(out_channel*4, out_channel*8),
            self.make_convpool_block(out_channel*8, out_channel*16))

        num_features_before_fcnn = functools.reduce(operator.mul,
                                                    self.feature_extractor(torch.rand(1,
                                                                                      in_channel,
                                                                                      input_dim)).shape)

        output_layers = [nn.Dropout(dropout),
                         nn.Linear(in_features=num_features_before_fcnn, out_features=20),
                         nn.BatchNorm1d(20),
                         nn.LeakyReLU(0.1),
                         nn.Linear(in_features=20, out_features=1)];

        if is_classifier:
            output_layers.append(nn.Sigmoid())

        self.output_block = nn.Sequential(*output_layers)

    def make_convpool_block(self, input_channels, output_channels, kernel_size=3, stride=1):
        return nn.Sequential(
            nn.Conv1d(input_channels, output_channels, kernel_size, bias=False),
            nn.BatchNorm1d(output_channels),
            nn.LeakyReLU(0.1),
            nn.AvgPool1d(3))

    def forward(self, x):
        batch_size = x.size(0)
        out = self.feature_extractor(x)
        out = out.view(batch_size, -1)  # flatten the vector
        out = self.output_block(out)
        return out

# Cell
def weights_init(m):
    if isinstance(m, nn.Conv1d) or isinstance(m, nn.Linear):
        torch.nn.init.kaiming_uniform_(m.weight)
    if isinstance(m, nn.BatchNorm1d):
        torch.nn.init.normal_(m.weight, 0.0, 0.02)
        torch.nn.init.constant_(m.bias, 0)

# Cell
class Learner():
    def __init__(self, model, criterion, opt, n_epochs=50,
                 scheduler=None, verbose=True):
        store_attr() # see https://fastpages.fast.ai/fastcore
        use_cuda = torch.cuda.is_available()
        self.device = torch.device('cuda:0' if use_cuda else 'cpu')
        self._init_losses()
        print(f'Runtime is: {self.device}')

    def fit(self, training_generator, validation_generator,
            losses_append=None, loss_valid_stop=-1, loss_train_stop=-1):
        if losses_append is None: self._init_losses()
        for epoch in tqdm(range(self.n_epochs)):
            loss_valid = []
            loss_train = []

            # Training loop
            self.model.train()
            for batches in training_generator:
                local_batch, local_labels = self._to_device(batches)
                self.opt.zero_grad() #
                y_hat, loss = self.eval(local_batch, local_labels)
                loss.backward(retain_graph=True) #
                self.opt.step() #
                loss_train.append(loss.item())


            # Validation loop
            self.model.eval()
            with torch.no_grad():
                for batches in validation_generator:
                    local_batch, local_labels = self._to_device(batches)
                    y_hat, loss = self.eval(local_batch, local_labels)
                    loss_valid.append(loss.item())

            # mean losses after n epoch
            mean_loss_valid = np.mean(np.array(loss_valid))
            mean_loss_train = np.mean(np.array(loss_train))

            if self.scheduler is not None:
                self.scheduler.step()

            if self.verbose:
                print(f'End of Epoch {epoch + 1}\n Training loss: {mean_loss_train}\n Validation loss: {mean_loss_valid}')

            self.losses['train'].append(mean_loss_train)
            self.losses['valid'].append(mean_loss_valid)

            if mean_loss_valid < loss_valid_stop:
                print('Goal achieved!')
                break
            if (mean_loss_train < loss_train_stop) and not (epoch % 10):
                print('Goal achieved!')
                break
        return self.model, self.losses

    def eval(self, X, y):
        y_hat = self.model(X)
        loss = self.criterion(y_hat, y)
        return y_hat, loss

    def predict(self, data_generator, sample=None):
        self.model.eval() if sample is None else self._MCDropoutOn()
        ys_hat, ys = [], []
        if sample is not None: raise Exception(f'MCDropout needs refactoring')
        with torch.no_grad():
            for batches in data_generator:
                local_batch, local_labels = self._to_device(batches)
                if sample is None:
                    ys_hat.append(self.model(local_batch))
                else:
                    # TO BE FIXED: MCDROPOUT
                    y_hat = np.stack([self.model(local_batch).cpu() for s in tqdm(range(sample))]).squeeze()
                ys.append(local_labels)
        return (self._npify(torch.cat(y)) for y in (ys_hat, ys))

    def _to_device(self, batches):
        if self.device.type == 'cpu': return batches
        return (batch.to(self.device) for batch in batches)

    def _init_losses(self):
        self.losses = {'train': [], 'valid': []}

    def _MCDropoutOn(self):
         # Check https://discuss.pytorch.org/t/turn-off-batch-norm-but-leave-dropout-on/14815
        # to deactivate BatchNorm during inference (maybe not critical when predicting all set)
        # and
        # for m in model.modules():
        #   if isinstance(m, nn.BatchNorm1d):
        #     m.eval()
        self.model.train()
        for m in self.model.modules():
            if isinstance(m, nn.BatchNorm1d):
                m.eval()
    def _npify(self, tensor):
        if self.device.type == 'cpu': return tensor.numpy()
        return tensor.cpu().numpy()


# Cell
class Evaluator():
    def __init__(self, model, data, depth_order,
                 seeds=range(20), device='cpu', n_epochs=5,
                 verbose=True, split_ratio=0.1):
        self.seeds = seeds
        self.model = model
        self.X, self.y = data
        self.in_shape = self.X.shape[1]
        self.depth_order = depth_order
        self.device = device
        self.split_ratio = split_ratio
        self.n_epochs = n_epochs
        self.verbose = verbose

        self.losses = []
        # self.models = []
        self.learners = []
        self.perfs = OrderedDict({'train': [], 'test': []})

    def train_multiple(self,
                       optimizer=torch.optim.Adam,
                       scheduler=torch.optim.lr_scheduler.CyclicLR,
                       criterion=nn.MSELoss(),
                       sc_kwargs={}):

        for seed in tqdm(self.seeds):
            X_train, X_test, y_train, y_test, depth_order_train, depth_order_test = self._splitter(seed)

            dls = DataLoaders(((X_train, y_train), (X_test, y_test)),
                              transform=SNV_transform())
            training_generator, validation_generator = dls.loaders()

            model = self.model(self.in_shape).to(self.device)
            opt = optimizer(model.parameters())
            model = model.apply(weights_init)

            learner = Learner(model, criterion, opt, n_epochs=self.n_epochs,
                              scheduler=scheduler(opt, **sc_kwargs), verbose=self.verbose)

            model, loss = learner.fit(training_generator, validation_generator)

            self.losses.append(loss)
            # self.models.append(model)
            self.learners.append(learner)

    def eval_on_train(self, reducer):
        perfs = []
        for i, seed in enumerate(self.seeds):
            X_train, X_test, y_train, y_test, _, _ = self._splitter(seed)

            dls = DataLoaders(((X_train, y_train), (X_test, y_test)),
                              transform=SNV_transform())
            train_generators , _ = dls.loaders()

            y_hat, y_true = self.learners[i].predict(train_generators)
            perf = eval_reg(y_true, y_hat)
            # perf = self.models[i].eval((X_train, y_train))
            perf['n'] = len(X_train)
            perfs.append(perf)
        if reducer:
            perfs = self.reduce(perfs, reducer)
        return perfs

    def eval_on_test(self, order=-1, reducer=None):
        perfs = []
        for i, seed in enumerate(self.seeds):
            X_train, X_test, y_train, y_test, depth_order_train, depth_order_test = self._splitter(seed)
            if order != - 1:
                mask = depth_order_test[:, 1] == order
                X_test, y_test = X_test[mask, :], y_test[mask]

            dls = DataLoaders(((X_train, y_train), (X_test, y_test)),
                              transform=SNV_transform())
            _, test_generators = dls.loaders()

            y_hat, y_true = self.learners[i].predict(test_generators)
            perf = eval_reg(y_true, y_hat)
            perf['n'] = len(X_test)
            perfs.append(perf)
        if reducer:
            perfs = self.reduce(perfs, reducer)
        return perfs

    def _splitter(self, seed):
        return train_test_split(self.X, self.y, self.depth_order,
                                test_size=self.split_ratio,
                                random_state=seed)

    def reduce(self, perfs, fn=np.mean):
        results = {}
        for metric in perfs[0].keys():
            result = fn(np.array([perf[metric] for perf in perfs]))
            results[metric] = result
        return results